#include <stdio.h>
#include <getopt.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/resource.h>

#include "gpio.h"
#include "serial.h"
#include "r8c-bits.h"

int verbose;
int readback = 0;
int zone = 0;
int console = 0;

#define ZRESET		zone_resets[zone]
#define ZP_MODE		61

static int zone_resets[] = { 64, 65, 66, 67, 68 };
static int zone_mux[] = { 2, 1, 0, 3, 4 };

static void mspause(int ms)
{
  struct timeval s, e;
  gettimeofday(&s, 0);
  /*printf("%3d %d.%06d : ", ms, s.tv_sec, s.tv_usec);*/
  s.tv_usec += ms * 1000;
  if (s.tv_usec > 999999)
    {
      s.tv_sec += s.tv_usec / 1000000;
      s.tv_usec %= 1000000;
    }
  while (1)
    {
      gettimeofday(&e, 0);
      if (e.tv_sec > s.tv_sec
	  || (e.tv_sec == s.tv_sec
	      && e.tv_usec >= s.tv_usec))
	{
	  /*printf("%d.%06d\n", e.tv_sec, e.tv_usec);*/
	  return;
	}
    }
}

static void
set_mux ()
{
  gpio_set_mode (58, GPIO_MODE_WRITE, GPIO_MODE_GPIO);
  gpio_set_mode (59, GPIO_MODE_WRITE, GPIO_MODE_GPIO);
  gpio_set_mode (60, GPIO_MODE_WRITE, GPIO_MODE_GPIO);
  gpio_write (58, zone_mux[zone] & 1);
  gpio_write (59, zone_mux[zone] & 2);
  gpio_write (60, zone_mux[zone] & 4);
}

static int id_d;

void
check_status (int verbose)
{
  static char *id_string[] = { "\033[31mNot ID'd",
			       "\033[33mID not matched",
			       "Res",
			       "\033[32mID'd" };
  int srd0, srd1;

  while (1)
    {
      serial_write (0x70);
      srd0 = serial_read ();
      if (srd0 == -1)
	exit (1);
      srd1 = serial_read ();
      if (srd1 == -1)
	exit (1);
      id_d = (srd1>>2) & 0x03;
      if (verbose)
	printf("Status: Seq: %s\033[0m  Erase: %s\033[0m  Program: %s\033[0m  ID: %s\033[0m\n",
	       srd0 & 0x80 ? "\033[32mReady" : "\033[34mBusy",
	       srd0 & 0x20 ? "\033[31mError" : "\033[32mNormal",
	       srd0 & 0x10 ? "\033[31mError" : "\033[32mNormal",
	       id_string[id_d]);
      if (srd0 & 0x80)
	break;
      mspause(1);
    };
}

void
clear_status()
{
  serial_write (0x50);
}

void
check_id ()
{
  printf("ID needed\n");
  exit (0);
}

void
program_page (int page, unsigned char *bytes)
{
  unsigned char buf[256];
  int i, j, tries=0;

  printf("\r%06x", page*256);
  fflush(stdout);
  while (tries < 3)
    {
      clear_status ();
      serial_write (0x41);
      serial_write (page % 256);
      serial_write (page / 256);
      for (i=0; i<256; i++)
	serial_write (bytes[i]);
      check_status (0);

      if (!readback)
	break;

      serial_write (0xff);
      serial_write (page % 256);
      serial_write (page / 256);
      for (i=0; i<256; i++)
	buf[i] = serial_read ();

      if (memcmp (buf, bytes, 256) == 0)
	break;

      printf("\r%06x: compare mismatch, trial %d\n", page*256, tries);
      for (i=0; i<256; i+=16)
	if (memcmp (buf+i, bytes+i, 16))
	  {
	    printf("\033[32mW %06x:", i+j);
	    for (j=0; j<16; j++)
	      printf(" %02x", bytes[i+j]);
	    printf("\n");
	    printf("\033[31mR %06x:", i+j);
	    for (j=0; j<16; j++)
	      printf(" %02x", buf[i+j]);
	    printf("\033[0m\n");
	    break;
	  }

      tries ++;
    }
  if (tries == 3)
    exit(1);
}

static int
try_unlocking (int byte)
{
  int i;
  clear_status ();
  serial_write (0xf5);
  serial_write (0xdf);
  serial_write (0xff);
  serial_write (0x00);
  serial_write (0x07);
  for (i=0; i<7; i++)
    serial_write (byte);
  check_status (1);
  return id_d == 3;
}

static void
usage ()
{
  fprintf (stderr,
	   "usage: r8c-prog [-v] program"
	   " program\n");
  exit (1);
}

int
main (int argc, char **argv)
{
  int i, o, j;
  char *fname;
  FILE *f;
  unsigned char buf[256];
  int pageno;

  setbuf (stdout, 0);

  while ((o = getopt (argc, argv, "z:vcr")) != -1)
    switch (o)
      {
      case 'v':
	verbose++;
	break;
      case 'r':
	readback++;
	break;
      case 'z':
	zone = atoi (optarg);
	if (zone < 1 || zone > 5)
	  {
	    printf("invalid zone %d must be 1..5\n");
	    exit(1);
	  }
	zone --;
	break;
      case 'c':
	console = 1;
	break;
      case '?':
	usage ();
      }

  fname = argv[optind];
  if (!fname)
    usage ();

  f = fopen(fname, "rb");
  if (!f)
    {
      perror(fname);
      exit(1);
    }
  fread (buf, 1, 5, f);
  if (memcmp (buf, "R8C\0", 4))
    {
      printf("Error: not an R8C page file\n");
      exit (1);
    }

  gpio_setup ();
/*serial_init ("/dev/ttyS2", 9600);*/
  serial_init ("/dev/ttyS2", 38400);

  if (setpriority (PRIO_PROCESS, 0, -20) == -1)
    perror("setpriority");

  gpio_set_mode (ZRESET, GPIO_MODE_WRITE, GPIO_MODE_GPIO);
  gpio_set_mode (ZP_MODE, GPIO_MODE_WRITE, GPIO_MODE_GPIO);
  set_mux ();

  /* RESET */
  gpio_write (ZRESET, 0);
  gpio_write (ZP_MODE, 0);
  mspause(10);

  /* STANDARD SERIAL MODE */
  gpio_write (ZRESET, 1);
  mspause (300);
  serial_drain ();

  /* BAUD RATE 9600 */
  for (i=0; i<16; i++)
    {
      serial_write (0);
      serial_sync ();
      mspause (20);
    }
  serial_drain ();
  j = 0;
  serial_set_timeout (100);
  do {
    serial_write (0xb0);
    i = serial_read ();
    printf("baud = 0x%02x\n", i);
    j++;
  } while (i != 0xb0 && j < 5);
  if (i != 0xb0)
    exit(1);

  serial_set_timeout (100);

  /* VERSION CHECK */
  serial_write (0xfb);
  printf("version = \"");
  for (i=0; i<8; i++)
    putchar(serial_read ());
  printf("\"\n");

  check_status(1);

  if (id_d != 3)
    try_unlocking (0);
  if (id_d != 3)
    try_unlocking (0xff);
  if (id_d != 3)
    check_id ();

  {
    unsigned char *p = r8cbits_map[0xff];
    printf("ID = %02x %02x %02x %02x %02x %02x %02x,  Watchdog = %02x\n",
	   p[0xdf], p[0xe3], p[0xeb],
	   p[0xef], p[0xf3], p[0xf7],
	   p[0xfb], p[0xff]);
  }

  serial_set_timeout (10000);

  /* ERASE ALL */
  clear_status ();
  serial_write (0xA7);
  serial_write (0xD0);
  serial_sync ();
  mspause(20);
  check_status(1);

  /* PROGRAM */
  while (1)
    {
      pageno = fgetc (f) & 0xff;
      if (feof (f))
	break;
      fread (buf, 1, 256, f);
      program_page (pageno, buf);
    }

  printf("\n");
  printf("Running...\n");
  gpio_write (ZP_MODE, 1);
  gpio_write (ZRESET, 0);
  mspause(1);
  gpio_write (ZRESET, 1);

  if (console)
    {
      int started = 0;
      char c;
      serial_set_timeout (0);
      printf("\n\033[34m-- console --\033[0m\n");
      while (1)
	{
	  c = serial_read();
	  if (c == 0 && !started)
	    continue;
	  started = 1;
	  if (isgraph(c) || c == '\r' || c == '\n' || c == ' ' || c == '\t' || c == 27)
	    fputc(c, stdout);
	  else
	    printf("\033[31m[%02x]\033[0m", (unsigned char)c);
	}
    }
}
